@@ -6,13 +6,15 @@
  */
 
 #include "mathutils/log_factorial_lookup_table.hpp"
+#include "mathutils/simple_generator.hpp"
 #include "mathutils/spherical_harmonics_index_lookup_table.hpp"
 #include <Eigen/Dense>
 #include <array>
 #include <cmath>
+#include <complex>
+#include <coroutine>
 #include <cstdint>
 #include <stdexcept>
-// #include <complex>
 
 /////////////////////////////////////
 /////////////////////////////////////
@@ -168,7 +170,8 @@ inline std::array<int, 2> spherical_harmonic_index_lm_N(int n) {
 /**
  * @brief Compute the associated Legendre polynomial with spherical harmonics
  * normalization \f$P_{m m}(\theta) = e^{-im\phi}Y_{m m}(\theta, \phi)\f$
- * for \f$0\leq\theta\leq\pi/2\f$ and \f$m\geq 0\f$.
+ * for \f$0\leq\theta\leq\pi/2\f$ and \f$m\geq 0\f$. Used to initialize the
+ * recursion.
  * @param m Degree/order of the polynomial (\f$m=0,1,\ldots\f$).
  * @param theta Polar angle in radians (\f$0\leq\theta\leq\pi/2\f$).
  * @return double Value of the associated normalized Legendre polynomial.
@@ -192,33 +195,44 @@ inline double reduced_spherical_Pmm(int m, double theta) {
                             0.5 * std::lgamma(2 * m + 2) - std::lgamma(m + 1));
 }
 
+// inline std::pair<double, double>
+// init_Plm_recursion_three_term_upward_ell(int m, double theta) {
+//   double P_m_m =
+//       (theta == 0.0)
+//           ? (m == 0) / std::sqrt(4 * M_PI)
+//           : (1 - 2 * (m & 1)) *
+//                 std::exp(m * std::log(std::sin(theta)) -
+//                          0.5 * std::log(4 * M_PI) - m * std::log(2) +
+//                          0.5 * std::lgamma(2 * m + 2) - std::lgamma(m + 1));
+//   double P_m_plus_one_m = std::sqrt(2 * m + 3) * std::cos(theta) * P_m_m;
+//   return {P_m_m, P_m_plus_one_m};
+// }
+
 /**
- * @brief Compute the associated Legendre polynomial with spherical harmonics
- * normalization \f$P_{\ell m}(\theta) = e^{-im\phi}Y_{\ell m}(\theta, \phi)\f$
- * for \f$0\leq\theta\leq\pi/2\f$ and \f$m\geq 0\f$.
- *
- * @param l Degree of the polynomial (\f$\ell=0,1,\ldots\f$).
- * @param m Order of the polynomial (\f$m=0,1,\ldots,\ell\f$).
- * @param theta Polar angle in radians (\f$0\leq\theta\leq\pi/2\f$).
- * @return double Value of the associated normalized Legendre polynomial.
+ * @brief Generate the associated Legendre polynomial with spherical harmonics
+ * normalization \f$P_{\ell m}(\theta) = e^{-im\phi}Y_{\ell m}(\theta,
+ * \phi)\f$ for \f$\ell = \ell_{start}, \ell_{start} + 1, \ldots, \ell_{end}\f$
+ * and \f$m=\ell_{start}\f$.
  */
-inline double reduced_spherical_Plm(int l, int m, double theta) {
-  double cos_theta = std::cos(theta);
-  // P_{m m}
-  double P_ell_minus_one_m = reduced_spherical_Pmm(m, theta);
-  if (l == m) {
-    return P_ell_minus_one_m;
+mathutils::SimpleGenerator<double>
+generate_reduced_spherical_Plm_recursion_three_term_upward_ell(int l_start,
+                                                               int l_end,
+                                                               double theta) {
+  if (l_start < 0 || l_end < l_start) {
+    co_return;
   }
-  // P_{m+1 m}
+  int m = l_start;
+  double P_ell_minus_one_m = reduced_spherical_Pmm(m, theta);
+  // ell = l_start
+  co_yield P_ell_minus_one_m;
+  if (l_start == l_end) {
+    co_return;
+  }
+  double cos_theta = std::cos(theta);
   double P_ell_m = cos_theta * std::sqrt(2 * m + 3) * P_ell_minus_one_m;
-  for (int ell = m + 2; ell <= l; ++ell) {
-    // double q =
-    //     cos_theta *
-    //         std::sqrt((2 * ell - 1) * (2 * ell + 1) / ((ell - m) * (ell +
-    //         m))) * P_ell_m -
-    //     std::sqrt((ell - m - 1) * (ell + m - 1) * (2 * ell + 1) /
-    //               ((ell - m) * (ell + m) * (2 * ell - 3))) *
-    //         P_ell_minus_one_m;
+  // ell = l_start + 1
+  co_yield P_ell_m;
+  for (int ell = m + 2; ell <= l_end; ++ell) {
     const double den_ell_m = (ell - m) * 1.0 * (ell + m);
     const double c_ell_m =
         std::sqrt(((2.0 * ell - 1.0) * (2.0 * ell + 1.0)) / den_ell_m);
@@ -231,6 +245,93 @@ inline double reduced_spherical_Plm(int l, int m, double theta) {
         cos_theta * c_ell_m * P_ell_m - c_ell_minus_one_m * P_ell_minus_one_m;
     P_ell_minus_one_m = P_ell_m;
     P_ell_m = q;
+    co_yield P_ell_m;
+  }
+}
+
+mathutils::SimpleGenerator<std::pair<int, double>>
+enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(int l_start,
+                                                                int l_end,
+                                                                double theta) {
+  if (l_start < 0 || l_end < l_start) {
+    co_return;
+  }
+  int m = l_start;
+  double P_ell_minus_one_m = reduced_spherical_Pmm(m, theta);
+  // ell = l_start
+  co_yield {l_start, P_ell_minus_one_m};
+  if (l_start == l_end) {
+    co_return;
+  }
+  double cos_theta = std::cos(theta);
+  double P_ell_m = cos_theta * std::sqrt(2 * m + 3) * P_ell_minus_one_m;
+  // ell = l_start + 1
+  co_yield {l_start + 1, P_ell_m};
+  for (int ell = m + 2; ell <= l_end; ++ell) {
+    const double den_ell_m = (ell - m) * 1.0 * (ell + m);
+    const double c_ell_m =
+        std::sqrt(((2.0 * ell - 1.0) * (2.0 * ell + 1.0)) / den_ell_m);
+    const double den_ell_minus_one_m =
+        (ell - m) * 1.0 * (ell + m) * (2.0 * ell - 3.0);
+    const double c_ell_minus_one_m =
+        std::sqrt(((ell - m - 1.0) * (ell + m - 1.0) * (2.0 * ell + 1.0)) /
+                  den_ell_minus_one_m);
+    const double q =
+        cos_theta * c_ell_m * P_ell_m - c_ell_minus_one_m * P_ell_minus_one_m;
+    P_ell_minus_one_m = P_ell_m;
+    P_ell_m = q;
+    co_yield {ell, P_ell_m};
+  }
+}
+
+/**
+ * @brief Compute the associated Legendre polynomial with spherical harmonics
+ * normalization \f$P_{\ell m}(\theta) = e^{-im\phi}Y_{\ell m}(\theta, \phi)\f$
+ * for \f$0\leq\theta\leq\pi/2\f$ and \f$m\geq 0\f$.
+ *
+ * @param l Degree of the polynomial (\f$\ell=0,1,\ldots\f$).
+ * @param m Order of the polynomial (\f$m=0,1,\ldots,\ell\f$).
+ * @param theta Polar angle in radians (\f$0\leq\theta\leq\pi/2\f$).
+ * @return double Value of the associated normalized Legendre polynomial.
+ */
+inline double reduced_spherical_Plm(int l, int m, double theta) {
+  // double cos_theta = std::cos(theta);
+  // // P_{m m}
+  // double P_ell_minus_one_m = reduced_spherical_Pmm(m, theta);
+  // if (l == m) {
+  //   return P_ell_minus_one_m;
+  // }
+  // // P_{m+1 m}
+  // double P_ell_m = cos_theta * std::sqrt(2 * m + 3) * P_ell_minus_one_m;
+  // for (int ell = m + 2; ell <= l; ++ell) {
+  //   // double q =
+  //   //     cos_theta *
+  //   //         std::sqrt((2 * ell - 1) * (2 * ell + 1) / ((ell - m) * (ell +
+  //   //         m))) * P_ell_m -
+  //   //     std::sqrt((ell - m - 1) * (ell + m - 1) * (2 * ell + 1) /
+  //   //               ((ell - m) * (ell + m) * (2 * ell - 3))) *
+  //   //         P_ell_minus_one_m;
+  //   const double den_ell_m = (ell - m) * 1.0 * (ell + m);
+  //   const double c_ell_m =
+  //       std::sqrt(((2.0 * ell - 1.0) * (2.0 * ell + 1.0)) / den_ell_m);
+  //   const double den_ell_minus_one_m =
+  //       (ell - m) * 1.0 * (ell + m) * (2.0 * ell - 3.0);
+  //   const double c_ell_minus_one_m =
+  //       std::sqrt(((ell - m - 1.0) * (ell + m - 1.0) * (2.0 * ell + 1.0)) /
+  //                 den_ell_minus_one_m);
+  //   const double q =
+  //       cos_theta * c_ell_m * P_ell_m - c_ell_minus_one_m *
+  //       P_ell_minus_one_m;
+  //   P_ell_minus_one_m = P_ell_m;
+  //   P_ell_m = q;
+  // }
+  // return P_ell_m;
+  double P_ell_m{0.0};
+  for (auto P : generate_reduced_spherical_Plm_recursion_three_term_upward_ell(
+           m, l, theta)) {
+    // ell_start = m
+    // ell_end = l
+    P_ell_m = P;
   }
   return P_ell_m;
 }
@@ -366,48 +467,115 @@ Eigen::VectorXd real_Ylm(int l, int m,
  * radians.
  * @return A matrix of shape (num_points, l_max * (l_max + 2) + 1)
  */
-// Eigen::MatrixXd compute_all_real_Ylm(int l_max,
-//                                      const Eigen::MatrixXd &thetaphi_coord_P)
-//                                      {
-//   // check l_max
-//   if (l_max < 0) {
-//     throw std::out_of_range("l_max must be non-negative");
-//   }
-//   // check thetaphi_coord_P
-//   if (thetaphi_coord_P.cols() != 2) {
-//     throw std::invalid_argument("thetaphi_coord_P must have 2 columns");
-//   }
-//   int num_points = thetaphi_coord_P.rows();
-//   if (num_points <= 0) {
-//     throw std::invalid_argument("thetaphi_coord_P must have at least one
-//     row");
-//   }
-//   //////////////////////////////////
-//   int num_modes = l_max * (l_max + 2) + 1; // = n_LM(l_max, l_max)+1
-//   Eigen::MatrixXd Y(num_points, num_modes);
-//   Y.setZero();
+Eigen::MatrixXd compute_all_real_Ylm(int l_max,
+                                     const Eigen::MatrixXd &thetaphi_coord_P) {
+  // check l_max
+  if (l_max < 0) {
+    throw std::out_of_range("l_max must be non-negative");
+  }
+  // check thetaphi_coord_P
+  if (thetaphi_coord_P.cols() != 2) {
+    throw std::invalid_argument("thetaphi_coord_P must have 2 columns");
+  }
+  int num_points = thetaphi_coord_P.rows();
+  if (num_points <= 0) {
+    throw std::invalid_argument("thetaphi_coord_P must have at least one row");
+  }
+  //////////////////////////////////
+  int num_modes = l_max * (l_max + 2) + 1; // = n_LM(l_max, l_max)+1
+  Eigen::MatrixXd Y(num_points, num_modes);
+  Y.setZero();
 
-//   for (int p{0}; p < num_points; ++p) {
-//     double theta = thetaphi_coord_P(p, 0);
-//     if (theta < 0 || theta > M_PI) {
-//       throw std::out_of_range("theta must be in the range [0, pi]");
-//     }
-//     double phi = thetaphi_coord_P(p, 1);
-//     for (int l{0}; l <= l_max; l++) {
-//       // m = 0 case
-//       int n = l * (l + 1);
-//       // ...
-//       for (int m{1}; m <= l; m++) {
-//         int n_plus = l * (l + 1) + m;
-//         int n_minus = l * (l + 1) - m;
+  for (int p{0}; p < num_points; ++p) {
+    double theta = thetaphi_coord_P(p, 0);
+    if (theta < 0 || theta > M_PI) {
+      throw std::out_of_range("theta must be in the range [0, pi]");
+    }
+    double phi = thetaphi_coord_P(p, 1);
+    double reduced_theta =
+        std::min(theta, M_PI - theta); // transform theta to [0, pi/2]
+    // m = 0
+    for (auto [l, reduced_P] :
+         enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(
+             0, l_max, reduced_theta)) {
+      int n = l * (l + 1);
+      Y(p, n) = ((theta > M_PI_2) && (l & 1)) ? -reduced_P : reduced_P;
+    }
+    // m = +/-1, ... , +/-l_max
+    for (int m = 1; m <= l_max; ++m) {
+      for (auto [l, reduced_P] :
+           enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(
+               m, l_max, reduced_theta)) {
+        int n_plus = l * (l + 1) + m;
+        int n_minus = l * (l + 1) - m;
+        // Q = (-1)^m * sqrt(2) * P_{l|m|}(theta)
+        double Q = ((m & 1) ^ ((theta > M_PI_2) && ((l - m) & 1)))
+                       ? -std::sqrt(2) * reduced_P
+                       : std::sqrt(2) * reduced_P;
+        // Y_{l |m|} = (-1)^m * sqrt(2) * P_{l|m|}*cos(|m|*phi)
+        Y(p, n_plus) = std::cos(m * phi) * Q;
+        // Y_{l, -|m|} = (-1)^m * sqrt(2) * P_{l|m|}*sin(|m|*phi)
+        Y(p, n_minus) = std::sin(m * phi) * Q;
+      }
+    }
+  } // end for p
+  return Y;
+}
 
-//         Y(p, n_minus) = std::sqrt(2) * std::sin(m * phi) * Y(p, n_plus);
-//         Y(p, n_plus) *= std::sqrt(2) * std::cos(m * phi);
-//       }
-//     }
-//   }
-//   return Y;
-// }
+Eigen::MatrixXcd compute_all_Ylm(int l_max,
+                                 const Eigen::MatrixXd &thetaphi_coord_P) {
+  // check l_max
+  if (l_max < 0) {
+    throw std::out_of_range("l_max must be non-negative");
+  }
+  // check thetaphi_coord_P
+  if (thetaphi_coord_P.cols() != 2) {
+    throw std::invalid_argument("thetaphi_coord_P must have 2 columns");
+  }
+  int num_points = thetaphi_coord_P.rows();
+  if (num_points <= 0) {
+    throw std::invalid_argument("thetaphi_coord_P must have at least one row");
+  }
+  //////////////////////////////////
+  int num_modes = l_max * (l_max + 2) + 1; // = n_LM(l_max, l_max)+1
+  Eigen::MatrixXcd Y(num_points, num_modes);
+  Y.setZero();
+
+  for (int p{0}; p < num_points; ++p) {
+    double theta = thetaphi_coord_P(p, 0);
+    if (theta < 0 || theta > M_PI) {
+      throw std::out_of_range("theta must be in the range [0, pi]");
+    }
+    double phi = thetaphi_coord_P(p, 1);
+    double reduced_theta =
+        std::min(theta, M_PI - theta); // transform theta to [0, pi/2]
+    // m = 0
+    for (auto [l, reduced_P] :
+         enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(
+             0, l_max, reduced_theta)) {
+      int n = l * (l + 1);
+      Y(p, n) = ((theta > M_PI_2) && (l & 1)) ? -reduced_P : reduced_P;
+    }
+    // m = +/-1, ... , +/-l_max
+    for (int m = 1; m <= l_max; ++m) {
+      for (auto [l, reduced_P] :
+           enumerate_reduced_spherical_Plm_recursion_three_term_upward_ell(
+               m, l_max, reduced_theta)) {
+        int n_plus = l * (l + 1) + m;
+        int n_minus = l * (l + 1) - m;
+        // Y_{l |m|} = P_{l|m|}*exp(i|m|phi)
+        Y(p, n_plus) =
+            ((theta > M_PI_2) && ((l - m) & 1))
+                ? -reduced_P * std::exp(std::complex<double>(0, m * phi))
+                : reduced_P * std::exp(std::complex<double>(0, m * phi));
+        // Y_{l, -|m|} = (-1)^m conj(Y_{l|m|})
+        Y(p, n_minus) =
+            (m & 1) ? -std::conj(Y(p, n_plus)) : std::conj(Y(p, n_plus));
+      }
+    }
+  } // end for p
+  return Y;
+}
 
 /**
  * @brief Compute the spherical harmonics \f$Y_{l m}(\theta, \phi)\f$
